module     olp_module
   implicit none
   private
   public :: OLP_Start, OLP_EvalSubProcess, OLP_Finalize, OLP_Option

contains

   subroutine     OLP_Start(contract_file_name,ierr) &
   & bind(C,name="OLP_Start")
      use, intrinsic :: iso_c_binding
      use p1_ddbar_neepscbar_matrix, only: p1_ddbar_neepscbar_initgolem => initgolem
      use p1_ddbar_neepscbar_config, only: p1_ddbar_neepscbar_PSP_rescue => PSP_rescue, &
           & p1_ddbar_neepscbar_PSP_verbosity => PSP_verbosity, &
           & p1_ddbar_neepscbar_PSP_chk_threshold1 => PSP_chk_threshold1, &
           & p1_ddbar_neepscbar_PSP_chk_threshold2 => PSP_chk_threshold2
      use p7_udbar_neepuubar_matrix, only: p7_udbar_neepuubar_initgolem => initgolem
      use p7_udbar_neepuubar_config, only: p7_udbar_neepuubar_PSP_rescue => PSP_rescue, &
           & p7_udbar_neepuubar_PSP_verbosity => PSP_verbosity, &
           & p7_udbar_neepuubar_PSP_chk_threshold1 => PSP_chk_threshold1, &
           & p7_udbar_neepuubar_PSP_chk_threshold2 => PSP_chk_threshold2
      use p0_ddbar_neepdubar_matrix, only: p0_ddbar_neepdubar_initgolem => initgolem
      use p0_ddbar_neepdubar_config, only: p0_ddbar_neepdubar_PSP_rescue => PSP_rescue, &
           & p0_ddbar_neepdubar_PSP_verbosity => PSP_verbosity, &
           & p0_ddbar_neepdubar_PSP_chk_threshold1 => PSP_chk_threshold1, &
           & p0_ddbar_neepdubar_PSP_chk_threshold2 => PSP_chk_threshold2
      use p8_udbar_neepssbar_matrix, only: p8_udbar_neepssbar_initgolem => initgolem
      use p8_udbar_neepssbar_config, only: p8_udbar_neepssbar_PSP_rescue => PSP_rescue, &
           & p8_udbar_neepssbar_PSP_verbosity => PSP_verbosity, &
           & p8_udbar_neepssbar_PSP_chk_threshold1 => PSP_chk_threshold1, &
           & p8_udbar_neepssbar_PSP_chk_threshold2 => PSP_chk_threshold2
      use p9_udbar_neepgg_matrix, only: p9_udbar_neepgg_initgolem => initgolem
      use p9_udbar_neepgg_config, only: p9_udbar_neepgg_PSP_rescue => PSP_rescue, &
           & p9_udbar_neepgg_PSP_verbosity => PSP_verbosity, &
           & p9_udbar_neepgg_PSP_chk_threshold1 => PSP_chk_threshold1, &
           & p9_udbar_neepgg_PSP_chk_threshold2 => PSP_chk_threshold2
      implicit none
      character(kind=c_char,len=1), intent(in) :: contract_file_name
      integer(kind=c_int), intent(out) :: ierr

      interface
         function strlen(s) bind(C,name='strlen')
            use, intrinsic :: iso_c_binding
            implicit none
            character(kind=c_char,len=1), intent(in) :: s
            integer(kind=c_int) :: strlen
         end function strlen
      end interface

      integer :: l, ferr
      character(len=128) :: line_buf
      character(len=9) :: kw
      integer :: PSP_verbosity, PSP_chk_threshold1, PSP_chk_threshold2
      logical :: PSP_rescue

      ierr = 1
      l = strlen(contract_file_name)

      open(unit=21, file=contract_file_name(1:l), &
          & status='old', action='read', iostat=ferr)

      if (ferr .ne. 0) then
         write(7,*) "In OLP_Start: ", contract_file_name(1:l), " not found!"
         ierr = -1
      end if

      do while (ferr .eq. 0)
         read(unit=21,fmt='(A128)',iostat=ferr) line_buf
         if (ferr .ne. 0) exit
         line_buf = adjustl(line_buf)
         kw = line_buf(1:9)
         do
            l = scan(kw, "DEFILMO")
            if (l .eq. 0) exit
            kw(l:l) = achar(ichar(kw(l:l)) - ichar('A') + ichar('a'))
         end do
         if (kw .eq. "modelfile") then
            line_buf = adjustl(line_buf(10:128))
            l = scan(line_buf, "|") - 1
            if(l .lt. 1) l = len(line_buf)
            l = len_trim(line_buf(1:l))
            exit
         end if
      end do

      close(unit=21)

      if (ierr .eq. 1) then
         call read_slha_file(line_buf(1:l))
      end if
      call p1_ddbar_neepscbar_initgolem(.true.)
      call p7_udbar_neepuubar_initgolem(.false.)
      call p0_ddbar_neepdubar_initgolem(.false.)
      call p8_udbar_neepssbar_initgolem(.false.)
      call p9_udbar_neepgg_initgolem(.false.)

      
      ! Uncomment to change rescue system setting on all suprocesses
      ! PSP_rescue = .true.
      ! PSP_verbosity = 2
      ! PSP_chk_threshold1 = 4
      ! PSP_chk_threshold2 = 3
      ! p1_ddbar_neepscbar_PSP_rescue = PSP_rescue
      ! p1_ddbar_neepscbar_PSP_verbosity =  PSP_verbosity
      ! p1_ddbar_neepscbar_PSP_chk_threshold1 = PSP_chk_threshold1
      ! p1_ddbar_neepscbar_PSP_chk_threshold2 = PSP_chk_threshold2
      ! p7_udbar_neepuubar_PSP_rescue = PSP_rescue
      ! p7_udbar_neepuubar_PSP_verbosity =  PSP_verbosity
      ! p7_udbar_neepuubar_PSP_chk_threshold1 = PSP_chk_threshold1
      ! p7_udbar_neepuubar_PSP_chk_threshold2 = PSP_chk_threshold2
      ! p0_ddbar_neepdubar_PSP_rescue = PSP_rescue
      ! p0_ddbar_neepdubar_PSP_verbosity =  PSP_verbosity
      ! p0_ddbar_neepdubar_PSP_chk_threshold1 = PSP_chk_threshold1
      ! p0_ddbar_neepdubar_PSP_chk_threshold2 = PSP_chk_threshold2
      ! p8_udbar_neepssbar_PSP_rescue = PSP_rescue
      ! p8_udbar_neepssbar_PSP_verbosity =  PSP_verbosity
      ! p8_udbar_neepssbar_PSP_chk_threshold1 = PSP_chk_threshold1
      ! p8_udbar_neepssbar_PSP_chk_threshold2 = PSP_chk_threshold2
      ! p9_udbar_neepgg_PSP_rescue = PSP_rescue
      ! p9_udbar_neepgg_PSP_verbosity =  PSP_verbosity
      ! p9_udbar_neepgg_PSP_chk_threshold1 = PSP_chk_threshold1
      ! p9_udbar_neepgg_PSP_chk_threshold2 = PSP_chk_threshold2

   end subroutine OLP_Start

   subroutine     OLP_EvalSubProcess(label, momenta, mu, parameters, res) &
   & bind(C,name="OLP_EvalSubProcess")
      use, intrinsic :: iso_c_binding
      implicit none
      integer(kind=c_int), value, intent(in) :: label
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(50), intent(in) :: momenta
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=c_double) :: alpha_s
      real(kind=c_double), parameter :: one_over_2pi = 0.15915494309189533577d0

      alpha_s = parameters(1)

      select case(label)
      case(0)
              call eval11(momenta(1:30), mu, parameters, res)
      case(1)
              call eval1(momenta(1:30), mu, parameters, res)
      case(2)
              call eval3(momenta(1:30), mu, parameters, res)
      case(3)
              call eval4(momenta(1:30), mu, parameters, res)
      case(4)
              call eval10(momenta(1:30), mu, parameters, res)
      case(5)
              call eval13(momenta(1:30), mu, parameters, res)
      case(6)
              call eval14(momenta(1:30), mu, parameters, res)
      case(7)
              call eval7(momenta(1:30), mu, parameters, res)
      case(8)
              call eval0(momenta(1:30), mu, parameters, res)
      case(9)
              call eval2(momenta(1:30), mu, parameters, res)
      case(10)
              call eval5(momenta(1:30), mu, parameters, res)
      case(11)
              call eval6(momenta(1:30), mu, parameters, res)
      case(12)
              call eval8(momenta(1:30), mu, parameters, res)
      case(13)
              call eval16(momenta(1:30), mu, parameters, res)
      case(14)
              call eval9(momenta(1:30), mu, parameters, res)
      case(15)
              call eval12(momenta(1:30), mu, parameters, res)
      case(16)
              call eval15(momenta(1:30), mu, parameters, res)
      case default
         res(:) = 0.0d0
      end select

      res(1:3) = alpha_s * one_over_2pi * res(1:3)
   end subroutine OLP_EvalSubProcess

   subroutine     OLP_Finalize() &
   & bind(C,name="OLP_Finalize")
      use, intrinsic :: iso_c_binding
      use p1_ddbar_neepscbar_matrix, only: p1_ddbar_neepscbar_exitgolem => exitgolem
      use p7_udbar_neepuubar_matrix, only: p7_udbar_neepuubar_exitgolem => exitgolem
      use p0_ddbar_neepdubar_matrix, only: p0_ddbar_neepdubar_exitgolem => exitgolem
      use p8_udbar_neepssbar_matrix, only: p8_udbar_neepssbar_exitgolem => exitgolem
      use p9_udbar_neepgg_matrix, only: p9_udbar_neepgg_exitgolem => exitgolem
      implicit none
      call p1_ddbar_neepscbar_exitgolem(.false.)
      call p7_udbar_neepuubar_exitgolem(.false.)
      call p0_ddbar_neepdubar_exitgolem(.false.)
      call p8_udbar_neepssbar_exitgolem(.false.)
      call p9_udbar_neepgg_exitgolem(.true.)
   end subroutine OLP_Finalize

   subroutine     OLP_Option(line,stat) &
   & bind(C,name="OLP_Option")
      use, intrinsic :: iso_c_binding
      use p1_ddbar_neepscbar_model, only: p1_ddbar_neepscbar_parseline => parseline
      use p7_udbar_neepuubar_model, only: p7_udbar_neepuubar_parseline => parseline
      use p0_ddbar_neepdubar_model, only: p0_ddbar_neepdubar_parseline => parseline
      use p8_udbar_neepssbar_model, only: p8_udbar_neepssbar_parseline => parseline
      use p9_udbar_neepgg_model, only: p9_udbar_neepgg_parseline => parseline
      implicit none
      character(kind=c_char,len=1), intent(in) :: line
      integer(kind=c_int), intent(out) :: stat
      integer :: l, ios

      interface
         function strlen(s) bind(C,name='strlen')
            use, intrinsic :: iso_c_binding
            implicit none
            character(kind=c_char,len=1), intent(in) :: s
            integer(kind=c_int) :: strlen
         end function strlen
      end interface

      l = strlen(line)
      call p1_ddbar_neepscbar_parseline(line(1:l),ios)
      if (ios .ne. 0) then
         stat = 0
         return
      end if
      call p7_udbar_neepuubar_parseline(line(1:l),ios)
      if (ios .ne. 0) then
         stat = 0
         return
      end if
      call p0_ddbar_neepdubar_parseline(line(1:l),ios)
      if (ios .ne. 0) then
         stat = 0
         return
      end if
      call p8_udbar_neepssbar_parseline(line(1:l),ios)
      if (ios .ne. 0) then
         stat = 0
         return
      end if
      call p9_udbar_neepgg_parseline(line(1:l),ios)
      if (ios .ne. 0) then
         stat = 0
         return
      end if
      stat = 1
   end subroutine OLP_Option
   !---#[ init_event_parameters :
   subroutine     init_event_parameters(sp, parameters)
      use, intrinsic :: iso_c_binding
      implicit none
      integer, intent(in) :: sp
      real(kind=c_double), dimension(10), intent(in) :: parameters
      !
      ! User hook for propagating scale dependent parameters to the
      ! model parameters in the subprocesses.
      !
      ! sp specifies the subprocess
      !
   end subroutine init_event_parameters
   !---#] init_event_parameters :

   !---#[ subroutine eval11 :
   subroutine     eval11(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p1_ddbar_neepscbar_config, only: ki
      use p1_ddbar_neepscbar_model, only: parseline
      use p11_cdbar_neepdbars_matrix, only: samplitude
      use p1_ddbar_neepscbar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(11, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval11
   !---#] subroutine eval11 :
   !---#[ subroutine eval1 :
   subroutine     eval1(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p1_ddbar_neepscbar_config, only: ki
      use p1_ddbar_neepscbar_model, only: parseline
      use p1_ddbar_neepscbar_matrix, only: samplitude
      use p1_ddbar_neepscbar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(1, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval1
   !---#] subroutine eval1 :
   !---#[ subroutine eval3 :
   subroutine     eval3(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p1_ddbar_neepscbar_config, only: ki
      use p1_ddbar_neepscbar_model, only: parseline
      use p3_dsbar_neepdcbar_matrix, only: samplitude
      use p1_ddbar_neepscbar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(3, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval3
   !---#] subroutine eval3 :
   !---#[ subroutine eval4 :
   subroutine     eval4(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p1_ddbar_neepscbar_config, only: ki
      use p1_ddbar_neepscbar_model, only: parseline
      use p4_dc_neepds_matrix, only: samplitude
      use p1_ddbar_neepscbar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(4, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval4
   !---#] subroutine eval4 :
   !---#[ subroutine eval10 :
   subroutine     eval10(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p7_udbar_neepuubar_config, only: ki
      use p7_udbar_neepuubar_model, only: parseline
      use p10_ubardbar_neepubarubar_matrix, only: samplitude
      use p7_udbar_neepuubar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(10, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval10
   !---#] subroutine eval10 :
   !---#[ subroutine eval13 :
   subroutine     eval13(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p7_udbar_neepuubar_config, only: ki
      use p7_udbar_neepuubar_model, only: parseline
      use p13_uu_neepud_matrix, only: samplitude
      use p7_udbar_neepuubar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(13, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval13
   !---#] subroutine eval13 :
   !---#[ subroutine eval14 :
   subroutine     eval14(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p7_udbar_neepuubar_config, only: ki
      use p7_udbar_neepuubar_model, only: parseline
      use p14_uubar_neepubard_matrix, only: samplitude
      use p7_udbar_neepuubar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(14, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval14
   !---#] subroutine eval14 :
   !---#[ subroutine eval7 :
   subroutine     eval7(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p7_udbar_neepuubar_config, only: ki
      use p7_udbar_neepuubar_model, only: parseline
      use p7_udbar_neepuubar_matrix, only: samplitude
      use p7_udbar_neepuubar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(7, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval7
   !---#] subroutine eval7 :
   !---#[ subroutine eval0 :
   subroutine     eval0(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p0_ddbar_neepdubar_config, only: ki
      use p0_ddbar_neepdubar_model, only: parseline
      use p0_ddbar_neepdubar_matrix, only: samplitude
      use p0_ddbar_neepdubar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(0, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval0
   !---#] subroutine eval0 :
   !---#[ subroutine eval2 :
   subroutine     eval2(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p0_ddbar_neepdubar_config, only: ki
      use p0_ddbar_neepdubar_model, only: parseline
      use p2_du_neepdd_matrix, only: samplitude
      use p0_ddbar_neepdubar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(2, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval2
   !---#] subroutine eval2 :
   !---#[ subroutine eval5 :
   subroutine     eval5(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p0_ddbar_neepdubar_config, only: ki
      use p0_ddbar_neepdubar_model, only: parseline
      use p5_dbardbar_neepdbarubar_matrix, only: samplitude
      use p0_ddbar_neepdubar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(5, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval5
   !---#] subroutine eval5 :
   !---#[ subroutine eval6 :
   subroutine     eval6(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p0_ddbar_neepdubar_config, only: ki
      use p0_ddbar_neepdubar_model, only: parseline
      use p6_udbar_neepddbar_matrix, only: samplitude
      use p0_ddbar_neepdubar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(6, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval6
   !---#] subroutine eval6 :
   !---#[ subroutine eval8 :
   subroutine     eval8(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p8_udbar_neepssbar_config, only: ki
      use p8_udbar_neepssbar_model, only: parseline
      use p8_udbar_neepssbar_matrix, only: samplitude
      use p8_udbar_neepssbar_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(8, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval8
   !---#] subroutine eval8 :
   !---#[ subroutine eval16 :
   subroutine     eval16(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p9_udbar_neepgg_config, only: ki
      use p9_udbar_neepgg_model, only: parseline
      use p16_gg_neepdubar_matrix, only: samplitude
      use p9_udbar_neepgg_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(16, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval16
   !---#] subroutine eval16 :
   !---#[ subroutine eval9 :
   subroutine     eval9(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p9_udbar_neepgg_config, only: ki
      use p9_udbar_neepgg_model, only: parseline
      use p9_udbar_neepgg_matrix, only: samplitude
      use p9_udbar_neepgg_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(9, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval9
   !---#] subroutine eval9 :
   !---#[ subroutine eval12 :
   subroutine     eval12(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p9_udbar_neepgg_config, only: ki
      use p9_udbar_neepgg_model, only: parseline
      use p12_gdbar_neepgubar_matrix, only: samplitude
      use p9_udbar_neepgg_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(12, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval12
   !---#] subroutine eval12 :
   !---#[ subroutine eval15 :
   subroutine     eval15(momenta, mu, parameters, res)
      use, intrinsic :: iso_c_binding
      use p9_udbar_neepgg_config, only: ki
      use p9_udbar_neepgg_model, only: parseline
      use p15_gu_neepgd_matrix, only: samplitude
      use p9_udbar_neepgg_groups, only: tear_down_golem95
      implicit none
      real(kind=c_double), dimension(30), intent(in) :: momenta
      real(kind=c_double), value, intent(in) :: mu
      real(kind=c_double), dimension(10), intent(in) :: parameters
      real(kind=c_double), dimension(4), intent(out) :: res

      real(kind=ki), dimension(6,4) :: vecs
      real(kind=ki), dimension(4) :: amp
      logical :: ok

      call init_event_parameters(15, parameters)

      vecs(:,1) = real(momenta(1::5),ki)
      vecs(:,2) = real(momenta(2::5),ki)
      vecs(:,3) = real(momenta(3::5),ki)
      vecs(:,4) = real(momenta(4::5),ki)

      call samplitude(vecs, mu*mu, amp, ok)
      call tear_down_golem95()

      if (ok) then
         !
      else
         !
      end if

      res(1) = real(amp(4), c_double)
      res(2) = real(amp(3), c_double)
      res(3) = real(amp(2), c_double)
      res(4) = real(amp(1), c_double)
   end subroutine eval15
   !---#] subroutine eval15 :

   subroutine     read_slha_file(line)
      use p1_ddbar_neepscbar_model, only: p1_ddbar_neepscbar_read_slha => read_slha
      use p7_udbar_neepuubar_model, only: p7_udbar_neepuubar_read_slha => read_slha
      use p0_ddbar_neepdubar_model, only: p0_ddbar_neepdubar_read_slha => read_slha
      use p8_udbar_neepssbar_model, only: p8_udbar_neepssbar_read_slha => read_slha
      use p9_udbar_neepgg_model, only: p9_udbar_neepgg_read_slha => read_slha
      implicit none
      character(len=*), intent(in) :: line
      character(len=512) :: file_name
      integer :: ierr

      call unescape_file_name(line, file_name)
      open(unit=27,file=file_name,status='old',iostat=ierr)
      if(ierr.ne.0) then
         print*, "Could not find SLHA model file"
      else
         call p1_ddbar_neepscbar_read_slha(27)
         rewind(unit=27)
         call p7_udbar_neepuubar_read_slha(27)
         rewind(unit=27)
         call p0_ddbar_neepdubar_read_slha(27)
         rewind(unit=27)
         call p8_udbar_neepssbar_read_slha(27)
         rewind(unit=27)
         call p9_udbar_neepgg_read_slha(27)
         close(27)
      end if
   end subroutine read_slha_file

   subroutine     unescape_file_name(source, dest)
      implicit none
      character(len=*), intent(in) :: source
      character(len=512), intent(out) :: dest
      integer :: is, id, l, hex, hexdigit, hexpos
      character(len=512) :: buf
      logical :: special

      is = scan(source, "|")

      if (is > 1) then
         buf = trim(source(1:is-1))
      else
         buf = trim(source)
      end if

      l = len(buf)
      id = 1
      special = .false.
      hexpos = 0
      if (buf(1:1) .eq. '"') then
         ! double quoted string
         do is = 2, l - 1
            if (special) then
               ! after a backslash or in \x.. escape
               if (hexpos == 1 .or. hexpos == 2) then
                  ! interpret hex digit
                  if ("0" .le. buf(is:is) .and. buf(is:is) .le. "9") then
                     hexdigit = ichar(buf(is:is)) - ichar("0")
                  elseif ("A" .le. buf(is:is) .and. buf(is:is) .le. "F") then
                     hexdigit = ichar(buf(is:is)) - ichar("A") + 10
                  elseif ("a" .le. buf(is:is) .and. buf(is:is) .le. "f") then
                     hexdigit = ichar(buf(is:is)) - ichar("a") + 10
                  else
                     print*, "Invalid hex escape sequence in file name"
                     stop
                  end if

                  if (hexpos == 1) then
                     hex = 16 * hexdigit
                     hexpos = 2
                  else
                     hex = hex + hexdigit
                     hexpos = 0
                     special = .false.
                     dest(id:id) = achar(hex)
                     id = id + 1
                  end if
               elseif (buf(is:is) .eq. "n") then
                  dest(id:id) = achar(10)
                  id = id + 1
                  special = .false.
               elseif (buf(is:is) .eq. "r") then
                  dest(id:id) = achar(13)
                  id = id + 1
                  special = .false.
               elseif (buf(is:is) .eq. "f") then
                  dest(id:id) = achar(12)
                  id = id + 1
                  special = .false.
               elseif (buf(is:is) .eq. "t") then
                  dest(id:id) = achar(9)
                  id = id + 1
                  special = .false.
               elseif (buf(is:is) .eq. "x") then
                  hexpos = 1
               else
                  dest(id:id) = buf(is:is)
                  id = id + 1
                  special = .false.
               end if
            else
               if(buf(is:is) .eq. "\") then
                  special = .true.
               else
                  dest(id:id) = source(is:is)
                  id = id + 1
               end if
            end if
         end do
      elseif (buf(1:1) .eq. '"') then
         ! single quoted string
         do is = 2, l - 1
            if (special) then
               dest(id:id) = buf(is:is)
               id = id + 1
               special = .false.
            elseif (buf(is:is) .eq. "'") then
               special = .true.
            else
               dest(id:id) = buf(is:is)
               id = id + 1
            end if
         end do
      else
         ! assume backslash escaped string
         do is = 1, l
            if (special) then
               dest(id:id) = buf(is:is)
               id = id + 1
               special = .false.
            elseif (buf(is:is) .eq. "\") then
               special = .true.
            else
               dest(id:id) = buf(is:is)
               id = id + 1
            end if
         end do
      end if
   end subroutine unescape_file_name
end module olp_module
